import torch
import numpy as np
import utils
import pathlib



# DataPoint
# preprocessed Repr of a Document for model (training, testing)

# m = 3 sentences in summaries
# p = 20 top scoring sentences
# k = 5 top scoring summaries

class DataPoint:

    def __init__(self, doc_str, sum_str, model_gensim, nlp): 

        # helper func for padding sentences to length 100
        def pad(seq, size = 100):
            if len(seq) < size:
                return torch.cat((seq, torch.zeros(size - len(seq), 200)), 0)
            return seq[:size]

        def preprocess(string):

            doc_preprocessed = nlp(string)
            
            doc_embeddings, raw_doc = [], []
            
            for sent in doc_preprocessed.sentences:
                
                sent_embeddings, raw_sent = [], []
                for tok in sent.tokens:
                    # raw output
                    raw_sent.append(tok.text) 

                    # get embeddings
                    try:
                        embedding = torch.Tensor(model_gensim[tok.text]) # output numpy array, convert to tensor, fix the warning?
                    except:
                        embedding = torch.zeros(200)
                    sent_embeddings.append(embedding)

                raw_doc.append(raw_sent)
                doc_embeddings.append(pad(torch.stack(sent_embeddings, 0)))
            document = torch.stack(doc_embeddings, 0)

            return raw_doc, document

        # preprocessed document
        self.raw_document, self.document = preprocess(doc_str)
        self.raw_summary, self.summary = preprocess(sum_str)
        

        # output of sentence encoder (not padded)
        self.sent_vecs = None
        self.gold_sent_vecs = None

        # Searchspace
        self.p_searchspace, self.n_searchspace, self.top_rouge, self.bin_summary = utils.searchspace(self.raw_document, self.raw_summary, 3)

    def compute_sent_vecs(self, model):
        # model.sentence_encoder
        # after training, switch model to m.eval()
 
        with torch.no_grad():
            self.sent_vecs = model.encode_sentences(self.document) #
            self.gold_sent_vecs = model.encode_sentences(self.summary)

# DataSet 
# assembles all datapoints, representing cnn_dailymail
class PreprocessedDataSet:
    def __init__(self, path, dataset=None, model_gensim=None, nlp=None): #arg path: Name der Partition (train, test, validation) oder Pfad (erstellt Ordner)
                                                                            # ['train0', ... 'train9']
        #self.path = pathlib.Path(path)
        self.path_list = [ pathlib.Path(p) for p in path ] 
        if dataset == None and model_gensim == None and nlp == None:
        
            #self.length = len(list(filter(lambda x:x.is_file(), self.path.iterdir()))) 
            lengths = []
            for p in self.path_list:
                lengths.append(len(list(filter(lambda x:x.is_file(), p.iterdir()))))
                        
            self.length = sum(lengths)
            self.length_list = lengths
            for i in range(1, len(lengths)):
                self.length_list[i] = self.length_list[i-1] + self.length_list[i] 

        elif dataset != None and model_gensim != None and nlp != None:

            self.path = path
        
            self.path.mkdir()
            
            self.length = len(dataset)
            
            for i, element in enumerate(dataset):
                article, highlights = element['article'], element['highlights']
                dp = DataPoint(article, highlights, model_gensim, nlp)
                torch.save(dp, self.path/str(i))
            
        else:
            raise Exception("INCONSISTENT ARGUMENTS")
    
    def __len__(self):
        return self.length
    
    def __getitem__(self, index):
    
        if index not in range(0, self.length):
            raise IndexError()
        
        try:
            # dp = torch.load(self.path/str(index))
            last_prefix = 0
            for i, prefix in enumerate(self.length_list): # prefix sum [100, 201, ..., len]
                
                if index < prefix:
                    new_index = index - last_prefix
                    # print(new_index)
                    break
                last_prefix = prefix

            dp = torch.load(self.path_list[i]/str(new_index))
                    

        except:
            raise FileNotFoundError()
        
        return dp
        
    def compute_sent_vecs(self, model, path):
        
        # writes in file, allows overwriting in main_ActorOnly
        pathlib.Path(path).mkdir(parents=True, exist_ok=True)

        for i in range(self.length):
            dp = self.__getitem__(i)
            dp.compute_sent_vecs(model)
            torch.save(dp, path/str(i))
    
    # Shorten dataset for testing
    def first(self, k):
        
        self.length = k

            








