from itertools import chain
from collections import Counter
import torch
import numpy as np
from quickselect import hoare
from itertools import combinations



def gaussian(x):

    return torch.exp(-x**2)


# helper func for selecting multiple elements from a data sequence in type list
def select_elements(iterable, idx_list):
    
    selection = []
    for i in idx_list:
        selection.append(iterable[i])
    
    return selection


# evaluation metric rouge
def rouge(summary, gold_summary, verbose=False):
    # takes two sequences of tokens in string, returns the mean value of rouge1, rouge2, rougeL, or the three separately if verbose = True

    def rouge_n(counts, gold_counts, length, gold_length):
        
        match = 0.0
        for tok in counts:
            match += min(counts[tok], gold_counts[tok])
        
        if length > 0:
            prec = match / length
        else:
            prec = 0
        if gold_length > 0:
            recall = match / gold_length
        else:
            recall = 0
        
        if prec == 0 or recall == 0:
            return 0
        
        return 2 * prec * recall / (prec + recall)
    
    def rouge_l(sents, gold_sents, counts, gold_counts, length, gold_length):
        
        lcs_scores = 0.0
        
        for sent in sents:
            token_mask = [0 for t in sent]
            sent_length = len(sent)
            for gold_sent in gold_sents:
                token_mask = lcs(sent, gold_sent, token_mask)
            
            cur_lcs_score = 0.0
            
            for i in range(sent_length):
                if token_mask[i]:
                    token = sent[i]
                    if counts[token] > 0 and gold_counts[token] > 0:
                        counts[token] -= 1
                        gold_counts[token] -= 1
                        cur_lcs_score += 1
            
            lcs_scores += cur_lcs_score

        if length > 0:
            prec = lcs_scores / length
        else:
            prec = 0

        if gold_length > 0:
            recall = lcs_scores / gold_length
        else:
            recall = 0
        
        if prec == 0 or recall == 0:
            return 0
        
        return 2 * prec * recall / (prec + recall)
    
    def lcs_grid(x, y):

        n = len(x)
        m = len(y)

        table = [[0 for i in range(m + 1)] for j in range(n + 1)]

        for j in range(m + 1):
            for i in range(n + 1):
                if i == 0 or j == 0:
                    cell = (0, 'e')
                elif x[i - 1] == y[j - 1]:
                    cell = (table[i - 1][j - 1][0] + 1, '\\')
                else:
                    over = table[i - 1][j][0]
                    left = table[i][j - 1][0]

                    if left < over:
                        cell = (over, '^')
                    else:
                        cell = (left, '<')

                table[i][j] = cell

        return table
    
    def lcs(x, y, mask_x):
        table = lcs_grid(x, y)
        i = len(x)
        j = len(y)

        while i > 0 and j > 0:
            move = table[i][j][1]
            if move == '\\':
                mask_x[i - 1] = 1
                i -= 1
                j -= 1
            elif move == '^':
                i -= 1
            elif move == '<':
                j -= 1

        return mask_x
    
    # counts unigrams, bigrams for rouge_n and rouge_l 
    summary = [[token.lower() for token in sent] for sent in summary]
    tokens = list(chain(*summary))
    
    gold_summary = [[token.lower() for token in sent] for sent in gold_summary]
    gold_tokens = list(chain(*gold_summary))
    
    unigram_counts = Counter(tokens)
    gold_unigram_counts = Counter(gold_tokens)
    
    bigram_counts = Counter(zip(tokens, tokens[1:]))
    gold_bigram_counts = Counter(zip(gold_tokens, gold_tokens[1:]))
    
    length_summary = len(tokens)
    length_gold_summary = len(gold_tokens)

    if verbose:
        return rouge_n(unigram_counts, gold_unigram_counts, length_summary, length_gold_summary), rouge_n(bigram_counts, gold_bigram_counts, length_summary - 1, length_gold_summary - 1), rouge_l(summary, gold_summary, unigram_counts, gold_unigram_counts, length_summary, length_gold_summary)
                
    return (rouge_n(unigram_counts, gold_unigram_counts, length_summary, length_gold_summary) + rouge_n(bigram_counts, gold_bigram_counts, length_summary - 1, length_gold_summary - 1) + rouge_l(summary, gold_summary, unigram_counts, gold_unigram_counts, length_summary, length_gold_summary)) / 3.0


# pre-compute searchspace of a document for high quality sampling in training
def searchspace(document, gold_summary, m, p=10, k=5):
    
    # select p highest scoring sentences of a document  
    if len(document) <= min(m, p):
        top_m_sents = list(range(len(document)))
        max_scoring_idx = list(range(len(document)))
    
    else:
        sent_idx = list(range(len(document)))
        sent_scores = list(map(lambda x:rouge([x], gold_summary), document))
        sent_key = lambda i:sent_scores[i]

        if len(document) <= m:
            top_m_sents = list(range(len(document)))
        else:
            hoare.nth_largest(sent_idx, m-1, key=sent_key)
            top_m_sents = sent_idx[:m]
            top_m_sents.sort()

        if len(document) <= p:
            max_scoring_idx = list(range(len(document)))
        else:
            hoare.nth_largest(sent_idx, p-1, key=sent_key)
            max_scoring_idx = sent_idx[:p]
            max_scoring_idx.sort()
    
    # define k (= 5) highest scoring summaries, compute p_searchspace, n_searchspace, top_rouge and bin_summary
    if len(max_scoring_idx) <= m:
        all_summaries = [tuple(max_scoring_idx)]
    else:
        all_summaries = list(combinations(max_scoring_idx, m))
    
    summary_idx = list(range(len(all_summaries)))
    summary_scores = list(map(lambda x:rouge(select_elements(document, x), gold_summary), all_summaries))
    summary_key = lambda i:summary_scores[i]
    
    if len(all_summaries) > k:
        hoare.nth_largest(summary_idx, k-1, key=summary_key)
        summary_idx = summary_idx[:k]
    
    max_scoring_summaries = select_elements(all_summaries, summary_idx)
    summary_scores = select_elements(summary_scores, summary_idx)
    
    p_searchspace = torch.zeros(len(max_scoring_summaries), len(document))
    n_searchspace = torch.ones(len(max_scoring_summaries), len(document))
    
    for i in range(len(max_scoring_summaries)):
        for idx in max_scoring_summaries[i]:
            p_searchspace[i,idx] = 1
            n_searchspace[i,idx] = 0
    
    top_rouge = torch.Tensor(summary_scores)

    bin_summary = torch.zeros(len(document))
    
    for idx in top_m_sents:
        bin_summary[idx] = 1

    return p_searchspace, n_searchspace, top_rouge, bin_summary




